package heapabstraction;
import java.io.PrintStream;

import com.ibm.wala.cfg.ShrikeCFG;
import com.ibm.wala.classLoader.IMethod;
import com.ibm.wala.fixpoint.FixedPointConstants;
import com.ibm.wala.fixpoint.IVariable;
import com.ibm.wala.shrikeBT.IInstruction;
import com.ibm.wala.shrikeBT.NewInstruction;
import com.ibm.wala.types.Descriptor;
import com.ibm.wala.types.TypeName;

import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.Stack;

/**
 * 
 * This class corresponds to the abstract state of your program.
 * There will be two AbstractStates for each node in your CFG, one 
 * for the entry and one for the exit. They will be updated by the 
 * Transfer Function.
 * 
 * 
 * @author asolar
 *
 */
class AbstractState implements IVariable<AbstractState>{
	int on=-1;		// order number
	int gnid=-1;	// what is this?
	
	// Environment variables.
	HashMap<Integer, HashSet<Integer>> sigmaBar;
	HashMap<Integer, HashMap<String, HashSet<Integer>>> hBar;
	Stack<HashSet <Integer>> operandStack;
	
	boolean isEntry = false;
	
	private static int allocationCount = 0;
	private static HashSet<Integer> registeredBlocks = new HashSet<Integer>(); 
	HashMap<IInstruction, Integer> instr2alloc;
	
	AbstractState(ShrikeCFG.BasicBlock n){
		// Initialize the abstract state based on the node.
		// Initialize entry state in special way.
		sigmaBar = new HashMap<Integer, HashSet<Integer>>();
		hBar = new HashMap<Integer, HashMap<String, HashSet<Integer>>>();
		
		// If we have an entry block, map the output variable names to
		// {null, out} in sigmaBar and hBar.
		if (n.isEntryBlock()) {
			isEntry = true;
			
			IMethod m = n.getMethod();
			Descriptor d = m.getDescriptor();
			TypeName[] types = d.getParameters();
			
			// Map this.
			sigmaBar.put(new Integer(0), getInitialMapping());
			initializeHeapFields(-1);
			
			for (int varCount = 0; varCount < types.length; ++varCount) {
				if (types[varCount].isPrimitiveType()) {
					sigmaBar.put(new Integer(varCount+1), getNullset());
				} else {
					// If it's not a primitive type, map in sigmaBar to {null, out}.
					sigmaBar.put(new Integer(varCount+1), getInitialMapping());
					initializeHeapFields(varCount+1);
					// Initialize current heap fields to {null, out}
				}
			}
			System.out.println("ENTRY BLOCK");
			System.out.println(sigmaBar);
			System.out.println(hBar);
		}
		
		// Initialize our mapping of instructions to allocation sites.
		instr2alloc = new HashMap<IInstruction, Integer>();
		if (n != null && n.getFirstInstructionIndex() >= 0) {
			Iterator<IInstruction> it = n.iterator();
			while (it.hasNext()) {
				IInstruction cur = it.next();
				if (cur instanceof NewInstruction) {
					instr2alloc.put(cur, allocationCount);
					if (!registeredBlocks.contains(new Integer(n.getFirstInstructionIndex()))) { 
						registeredBlocks.add(new Integer(n.getFirstInstructionIndex()));
						++allocationCount;
					}
				}
			}
		}
		
		operandStack = new Stack<HashSet <Integer>>();
	}
	
	public HashSet<Integer> getNullset() {
		HashSet<Integer> result = new HashSet<Integer>();
		result.add(null);
		return result;
	}
	
	public HashSet<Integer> getInitialMapping() {
		HashSet<Integer> initialMapping = new HashSet<Integer>();
		initialMapping.add(-1);
		initialMapping.add(null);
		return initialMapping;
	}
	
	public void initializeHeapFields(int loc) {
		if (isEntry) {
			Integer curLoc = new Integer(loc);
			if (hBar.get(curLoc) != null) {
				for (String curField : hBar.get(curLoc).keySet()) {
					setInHBar(curLoc, curField, getInitialMapping());
				}
			}
		}
	}
	
	public void setOperandStack(Stack<HashSet <Integer>> s) {
		operandStack = s;
	}
	public Stack<HashSet <Integer>> getOperandStack() {
		return operandStack;
	}
	public Stack<HashSet <Integer>> getOperandStackCopy() {
		return (Stack<HashSet <Integer>>)operandStack.clone();
	}
	
	private Stack<HashSet <Integer>> reverseStack(Stack<HashSet <Integer>> s) {
		Stack<HashSet <Integer>> result = new Stack<HashSet <Integer>>();
		while (!s.empty()) {
			result.push(s.pop());
		}
		return result;
	}
	public boolean mergeStacks(AbstractState rhs) {
		Stack<HashSet <Integer>> result = new Stack<HashSet <Integer>>();
		Stack<HashSet <Integer>> other = (Stack <HashSet<Integer>>)rhs.getOperandStack().clone();
		
		while (!operandStack.empty() && !other.empty()) {
			HashSet<Integer> cur = operandStack.pop();
			HashSet<Integer> otherCur = other.pop();
			if (otherCur != null) {
				cur.addAll(otherCur);
			}
			result.push(cur);
		}
		while (!operandStack.empty()) {
			result.push(operandStack.pop());
		}
		while (!other.empty()) {
			result.push((HashSet<Integer>)other.pop());
		}
		
		Stack<HashSet <Integer>> rev_result = reverseStack(result);
		setOperandStack(rev_result);
		return rev_result.equals(operandStack);
	}
	
	public boolean sameState(Object o) {
		if (o instanceof AbstractState) {
			AbstractState os = (AbstractState)o;
			return
				(sigmaBar.equals(os.getSigmaBar()) && hBar.equals(os.getHBar()));
		} else {
			return false;
		}
	}
	
	public void copyState(AbstractState as){
		on = as.getOrderNumber();
		gnid = as.getGraphNodeId();
		sigmaBar = (HashMap<Integer, HashSet<Integer>>) as.getSigmaBar().clone();
		hBar = (HashMap<Integer, HashMap<String, HashSet<Integer>>>)as.getHBar().clone();
	}
	
	// What do we use order number for?
	public int getOrderNumber(){
		return on;
	}
	public void setOrderNumber(int i){
		on = i;
	}
	
	@Override
	public int getGraphNodeId() {				
		return gnid;
	}

	@Override
	public void setGraphNodeId(int number) {
		gnid = number;		
	}
	
	public HashSet<Integer> getAllocSiteSet(IInstruction i) {
		HashSet<Integer> result = new HashSet<Integer>();
		result.add(instr2alloc.get(i));
		return result;
	}
	
	/* Sigma */
	public HashMap<Integer, HashSet<Integer>> getSigmaBar() {
		return sigmaBar;
	}
	public HashSet<Integer> lookupInSigmaBar(Integer loc) {
		HashSet<Integer> result = sigmaBar.get(loc);
		if (result == null) {
			result = new HashSet<Integer>();
			sigmaBar.put(loc, result);
			return result;
		} else {
			return result;
		}
	}
	public boolean addAllToSigmaBar(Integer loc, HashSet<Integer> vs) {
		assert vs != null : "Adding non-null to sigmaBar";
		HashSet<Integer> vals = (HashSet<Integer>)vs.clone();
		
		if (sigmaBar.get(loc) == null) {
			sigmaBar.put(loc, vals);
			return true;
		} else {
			return sigmaBar.get(loc).addAll(vals);
		}
	}
	public void setInSigmaBar(Integer loc, Integer val) {
		assert val != null : "Adding null to sigmaBar";
		HashSet<Integer> valset = new HashSet<Integer>();
		valset.add(val);
		sigmaBar.put(loc, valset);	
	}

	/* Heap */
	public HashMap<Integer, HashMap<String, HashSet<Integer>>> getHBar () {
		return hBar;
	}
	public HashSet<Integer> lookupInHBar(Integer loc, String field) {
		HashSet<Integer> result;
		
		if (hBar.get(loc) == null) {
			result = new HashSet<Integer>();
			if (loc != null && loc.equals(new Integer(-1))) {
				result.addAll(getInitialMapping());
			} else {
				result.add(null);
			}
			HashMap<String, HashSet<Integer>> result_map = new HashMap<String, HashSet<Integer>>();
			result_map.put(field, result);
			hBar.put(loc, result_map);
			return result;
		} else {
			result = (hBar.get(loc)).get(field);
			if (result == null) {
				result =  new HashSet<Integer>();
				if (loc != null && loc.equals(new Integer(-1))) {
					result.addAll(getInitialMapping());
				} else {
					result.add(null);
				}
				(hBar.get(loc)).put(field, result);
				return result;
			} else {
				return result;
			}
		}
	}
	public boolean addAllToHBar(Integer loc, String field, HashSet<Integer> v) {
		System.out.println("Adding to HBar[" + loc + ", " + field + "]: " + v);
		HashSet<Integer> vals = (HashSet<Integer>)v.clone();
		// vals.addAll(getInitialMapping());
		
		HashMap<String, HashSet<Integer>> fieldInfo = hBar.get(loc);
		if (fieldInfo == null) {
			if (loc != null && loc.equals(new Integer(-1))) {
				vals.addAll(getInitialMapping());
			} else {
				vals.add(null);
			}
			HashMap<String, HashSet<Integer>> newFields = new HashMap<String, HashSet<Integer>>();
			newFields.put(field, vals);
			hBar.put(loc, newFields);
			return true;
		} else {
			if (fieldInfo.get(field) == null) {
				if (loc != null && loc.equals(new Integer(-1))) {
					vals.addAll(getInitialMapping());
				} else {
					vals.add(null);
				}
				fieldInfo.put(field, vals);
				return true;
			} else {
				return fieldInfo.get(field).addAll(vals);
			}
		}
	}
	public void setInHBar(Integer loc, String field, HashSet<Integer> v) {
		HashSet<Integer> vals = (HashSet<Integer>)v.clone();
		
		HashMap<String, HashSet<Integer>> fieldInfo = hBar.get(loc);
		if (fieldInfo == null) {
			HashMap<String, HashSet<Integer>> newFields = new HashMap<String, HashSet<Integer>>();
			newFields.put(field, vals);
			hBar.put(loc, newFields);
		} else {
			fieldInfo.put(field, vals);
		}
	}
	
	
	/* Combine two abstract states and return whether they have changed. */
	public boolean combineChanged(AbstractState rhs) {
		System.out.println("Merging states!");
		// Now compare the input and output.
		if (sameState(rhs)) {
			return false;
		} else {
			boolean changed = false;
			HashMap<Integer, HashSet<Integer>> sigmaBarL = sigmaBar;
			HashMap<Integer, HashMap<String, HashSet<Integer>>> hBarL = hBar;
			HashMap<Integer, HashSet<Integer>> sigmaBarR = rhs.getSigmaBar();
			HashMap<Integer, HashMap<String, HashSet<Integer>>> hBarR = rhs.getHBar();
			
			// Update sigmaBar of the LHS.
			for (Integer rhsVar : sigmaBarR.keySet()) {
				if (sigmaBarR.get(rhsVar) != null) {
					changed = addAllToSigmaBar(rhsVar, sigmaBarR.get(rhsVar)) || changed;
				}
			} //rof sigmaBar
			
			// Update hBar of the LHS.
			for (Integer rhsHVar : hBarR.keySet()) {
				HashMap<String, HashSet<Integer>> hBarRLoc = hBarR.get(rhsHVar);
				for (String rhsHField : hBarRLoc.keySet()) {
					// Otherwise look up the set of locations associated with
					// (rhsHVar, rhsHField.)
					if (hBarRLoc.get(rhsHField) != null) {
						changed = addAllToHBar(rhsHVar, rhsHField, hBarRLoc.get(rhsHField)) || changed;
					}
				}
			}
			return changed;
		}
	}
	
	private void printHashSet(PrintStream p, HashSet<Integer> l) {
		p.print('{');
		
		Iterator<Integer> it = l.iterator();
		while (it.hasNext()) {
			p.print(it.next());
			if (it.hasNext()) {
				p.print(", ");
			}
		}
		p.print('}');
	}
	
	public void print(PrintStream p){
		// Iterate over sigmaBar.
		// varid	{ l_1, l_2, ..., l_k }
		// varid	{ l_1, l_2, ..., l_k }
		for (Integer key : sigmaBar.keySet()) {
			p.print(key);
			p.print('\t');
			HashSet<Integer> curHashSet = sigmaBar.get(key);
			printHashSet(p, curHashSet);
			p.print('\n');
		}
		
		// Iterate over the heap.
		// H[l1, f2]	{ l_1, l_2, ..., l_k }
		// H[lk, f2]	{ l_1, l_2, ..., l_k }
		for (Integer key : hBar.keySet()) {
			HashMap<String, HashSet<Integer>> curFields = hBar.get(key);
			for (String field : curFields.keySet()) {
				p.print("H[");
				p.print(key);
				p.print(", ");
				p.print(field);
				p.print("]\t");
				HashSet<Integer> curHashSet = curFields.get(field);
				printHashSet(p, curHashSet);
				p.print('\n');
			}
		}
	}
}

	    