package heapabstraction;
import java.io.File;
import java.io.IOException;
import java.util.Collection;
import java.util.Iterator;
import java.util.Properties;
import java.util.Set;
import java.util.jar.JarFile;

import com.ibm.wala.cfg.ShrikeCFG;
import com.ibm.wala.classLoader.IClass;
import com.ibm.wala.classLoader.IMethod;
import com.ibm.wala.classLoader.JarFileModule;
import com.ibm.wala.classLoader.ShrikeCTMethod;
import com.ibm.wala.client.AbstractAnalysisEngine;
import com.ibm.wala.dataflow.graph.BasicFramework;
import com.ibm.wala.ipa.callgraph.AnalysisCache;
import com.ibm.wala.ipa.callgraph.AnalysisOptions;
import com.ibm.wala.ipa.callgraph.CallGraphBuilder;
import com.ibm.wala.ipa.callgraph.Entrypoint;
import com.ibm.wala.ipa.callgraph.impl.Util;
import com.ibm.wala.ipa.cha.ClassHierarchyException;
import com.ibm.wala.ipa.cha.IClassHierarchy;
import com.ibm.wala.properties.WalaProperties;
import com.ibm.wala.types.TypeName;
import com.ibm.wala.types.TypeReference;
import com.ibm.wala.util.CancelException;
import com.ibm.wala.util.collections.HashSetFactory;
import com.ibm.wala.util.warnings.WalaException;


public class HeapAnalysis extends AbstractAnalysisEngine {

	private final Set<JarFile> applicationJarFiles;

	  private final String applicationMainClass;

	  /**
	   * The two input parameters define the program to analyze: the jars of .class files and the main class to start from.
	   */
	  public HeapAnalysis(Set<JarFile> applicationJarFiles, String applicationMainClass) {
	    this.applicationJarFiles = applicationJarFiles;
	    this.applicationMainClass = applicationMainClass;
	  }
	  
	/**
	   * Take the given set of JarFiles that constitute the program, and return a set of Module files as expected by the
	   * WALA machinery.
	   */
	  private Set<JarFileModule> getModuleFiles() {
	    Set<JarFileModule> result = HashSetFactory.make();
	    for (Iterator<JarFile> jars = applicationJarFiles.iterator(); jars.hasNext();) {
	      result.add(new JarFileModule(jars.next()));
	    }

	    return result;
	  }
	  
	  protected CallGraphBuilder getCallGraphBuilder(IClassHierarchy cha, AnalysisOptions options, AnalysisCache cache) {
		 // return Util.makeZeroContainerCFABuilder(options, cache, cha, scope);
		  //return Util.makeVanillaZeroOneCFABuilder(options, cache, cha, scope);
		   return Util.makeZeroCFABuilder(options, cache, cha, scope);
		  }
	  
	  
	  
	  
	  /**
	   * Collect the set of JarFiles that constitute the system libraries of the running JRE.
	   */
	  private JarFile[] getSystemJars() throws IOException {
	    String javaHomePath = "garbage";
	    Set<JarFile> jarFiles = HashSetFactory.make();

	    // first, see if wala.properties has been set up
	    try {
	      Properties p = WalaProperties.loadProperties();
	      javaHomePath = p.getProperty(WalaProperties.J2SE_DIR);
	    } catch (WalaException e) {
	      // no luck.
	    }

	    // if not, try assuming the running JRE looks normal
	    File x = new File(javaHomePath);
	    if (!(x.exists() && x.isDirectory())) {
	      javaHomePath = System.getProperty("java.home");

	      if (!javaHomePath.endsWith(File.separator)) {
	        javaHomePath = javaHomePath + File.separator;
	      }

	      javaHomePath = javaHomePath + "lib";
	    }

	    // find jars from chosen JRE lib path
	    collectJars(new File(javaHomePath), jarFiles);

	    return jarFiles.toArray(new JarFile[jarFiles.size()]);
	  }

	  /**
	   * Given a root path, add it to the set if it is a jar, or traverse it recursively if it is a directory.
	   */
	  private void collectJars(File f, Set<JarFile> result) throws IOException {
	    if (f.isDirectory()) {
	      File[] files = f.listFiles();
	      for (int i = 0; i < files.length; i++) {
	        collectJars(files[i], result);
	      }
	    } else if (f.getAbsolutePath().endsWith(".jar")) {
	      result.add(new JarFile(f));
	    }
	  }
	  
	  

	  
	  
	  Collection<IMethod> getMethodsforClass(String className, IClassHierarchy cha){
		  TypeReference T = TypeReference.findOrCreate(getScope().getApplicationLoader(), TypeName.string2TypeName(className));
		  IClass mclass = cha.lookupClass(T);
		  assert mclass != null : ("class " + className + " not found!");
		  // This is where it dies...
		  return mclass.getAllMethods();
	  }
	  

	  
	  ShrikeCFG getCFGforMethod(String className, String methodName, IClassHierarchy cha) throws CancelException{
		  Collection<IMethod> mset = getMethodsforClass(className, cha);
		  assert mset != null : ("methods for class " + className + " were null");
		  for(IMethod m : mset){
			  String mn = m.getName().toString();
			  if(mn.equals(methodName)){
				 	ShrikeCTMethod sm = (ShrikeCTMethod) m;
			    	ShrikeCFG cfg = new ShrikeCFG(sm);
			    	return cfg;
			  }
		  }
		  return null;
	  }
	  
	  ShrikeCFG getCFGforMain(String className, IClassHierarchy cha) throws CancelException{
		  
		  Iterable<Entrypoint> roots = Util.makeMainEntrypoints(getScope(), cha, applicationMainClass);
		  
		    Entrypoint e = roots.iterator().next();
		    
		    
	    	System.out.println(e);
	    	IMethod m = e.getMethod();
		  		
	    	ShrikeCTMethod sm = (ShrikeCTMethod) m;
	    	
	    	ShrikeCFG cfg = new ShrikeCFG(sm);
	    	return cfg;
	  }
	  
	  
	  
	  
	  void initialize()throws IOException, ClassHierarchyException, IllegalArgumentException,
      CancelException {

		    //
		    // set the application to analyze
		    //
		    setModuleFiles(getModuleFiles());

		    //
		    // set the system jar files to use.
		    // change this if you want to use a specific jre version
		    //
		    setJ2SELibraries(getSystemJars());


		    //
		    // the application and libraries are set, now build the scope...
		    //
		    buildAnalysisScope();

		    //
		    // ...and the class hierarchy
		    //
		  
	  }
	  
	  
	  
	  /**
	   * The heart of the analysis.
	   * @throws CancelException
	   * @throws IllegalArgumentException
	   */
	  public void mainRun(String methodName) throws IOException, ClassHierarchyException, IllegalArgumentException,
	      CancelException {
		initialize();
	    IClassHierarchy cha = buildClassHierarchy();
	    assert cha != null : "failed to create class hierarchy";
	    setClassHierarchy(cha);
	    
	    ShrikeCFG cfg = getCFGforMethod(applicationMainClass, methodName, cha); 
	    
	    System.out.println("CFG = " + cfg);
	    
	    TransferFunctionProvider functions = new TransferFunctionProvider();
	   
	    // Creates a new BasicFramework--this sets the TransferFunctionProvider.
	    BasicFramework<ShrikeCFG.BasicBlock, AbstractState> bf = new BasicFramework<ShrikeCFG.BasicBlock, AbstractState>(cfg, functions);
	    
	    // Creates a HeapAbstractionSolver.
	    HeapAbstractionSolver solver = new HeapAbstractionSolver(bf);
	    
	    solver.solve(null);
	    
	    solver.printExitState(System.out);
	  }

}
