package heapabstraction;
import java.util.HashSet;
import java.util.Iterator;
import java.util.Stack;

import com.ibm.wala.analysis.stackMachine.AbstractIntStackMachine.MachineState;
import com.ibm.wala.cfg.ShrikeCFG;
import com.ibm.wala.cfg.ShrikeCFG.BasicBlock;
import com.ibm.wala.classLoader.IBytecodeMethod;
import com.ibm.wala.fixedpoint.impl.UnaryOperator;
import com.ibm.wala.fixpoint.FixedPointConstants;
import com.ibm.wala.shrikeBT.ArrayLengthInstruction;
import com.ibm.wala.shrikeBT.CheckCastInstruction;
import com.ibm.wala.shrikeBT.ConstantInstruction;
import com.ibm.wala.shrikeBT.DupInstruction;
import com.ibm.wala.shrikeBT.GotoInstruction;
import com.ibm.wala.shrikeBT.IArrayLoadInstruction;
import com.ibm.wala.shrikeBT.IArrayStoreInstruction;
import com.ibm.wala.shrikeBT.IBinaryOpInstruction;
import com.ibm.wala.shrikeBT.IComparisonInstruction;
import com.ibm.wala.shrikeBT.IConditionalBranchInstruction;
import com.ibm.wala.shrikeBT.IConversionInstruction;
import com.ibm.wala.shrikeBT.IGetInstruction;
import com.ibm.wala.shrikeBT.IInstruction;
import com.ibm.wala.shrikeBT.IInvokeInstruction;
import com.ibm.wala.shrikeBT.ILoadIndirectInstruction;
import com.ibm.wala.shrikeBT.ILoadInstruction;
import com.ibm.wala.shrikeBT.IPutInstruction;
import com.ibm.wala.shrikeBT.IShiftInstruction;
import com.ibm.wala.shrikeBT.IStoreIndirectInstruction;
import com.ibm.wala.shrikeBT.IStoreInstruction;
import com.ibm.wala.shrikeBT.IUnaryOpInstruction;
import com.ibm.wala.shrikeBT.InstanceofInstruction;
import com.ibm.wala.shrikeBT.MonitorInstruction;
import com.ibm.wala.shrikeBT.NewInstruction;
import com.ibm.wala.shrikeBT.PopInstruction;
import com.ibm.wala.shrikeBT.ReturnInstruction;
import com.ibm.wala.shrikeBT.SwapInstruction;
import com.ibm.wala.shrikeBT.SwitchInstruction;
import com.ibm.wala.shrikeBT.ThrowInstruction;
import com.ibm.wala.shrikeBT.Util;
import com.ibm.wala.shrikeCT.InvalidClassFileException;
import com.ibm.wala.types.ClassLoaderReference;
import com.ibm.wala.types.TypeReference;
import com.ibm.wala.util.debug.Assertions;
import com.ibm.wala.util.shrike.ShrikeUtil;

class TransferFunction extends UnaryOperator<AbstractState>{
	class InstructionVisitor extends InstanceofInstruction.Visitor{
		AbstractState curState;
		Stack<HashSet <Integer>> operandStack;
		
		public InstructionVisitor(AbstractState as) {
			curState = as;
			operandStack = as.getOperandStack();
		}
		
		HashSet<Integer> nullset() {
			HashSet<Integer> nset = new HashSet<Integer>();
			nset.add(null);
			return nset;
		}
		HashSet<Integer> bottom() {
			return new HashSet<Integer>();
		}
		
		public HashSet<Integer> pop() {
			if (operandStack.empty()) {
				return new HashSet<Integer>();
			} else {
				return operandStack.pop();
			}
		}
		
		/**
		 * 
		 * You should modify this class so that instead of just 
		 * printing the instructions it actually updates the 
		 * abstract state appropriately. 
		 * You won't need to implement all the Instructions.
		 * Only implement as many as you need in order to 
		 * successfully analyze all the test methods in TheMainClass. 
		 * 
		 */
	    public void visitConstant(ConstantInstruction instruction) {
	    	operandStack.push(nullset());
	    	System.out.println("CONSTANT");
	    	System.out.println(instruction);
	    }

	    public void visitGoto(GotoInstruction instruction) {
	    	// Do nothing.
	    	System.out.println("GOTO");
	    	System.out.println(instruction);
	    }

	    public void visitLocalLoad(ILoadInstruction instruction) {
	    	HashSet<Integer> result = curState.lookupInSigmaBar(new Integer(instruction.getVarIndex()));
	    	System.out.println("LOCAL LOAD: " + instruction.getVarIndex() + "; got " + result);
	    	operandStack.push(result);
	    	System.out.println(instruction);
	    }

	    public void visitLocalStore(IStoreInstruction instruction) {
	        int index = instruction.getVarIndex();
	        HashSet<Integer> result = pop();
	        curState.addAllToSigmaBar(new Integer(index), result);
	        System.out.println("LOCAL STORE: " + index + " = " + result);
	    	System.out.println(instruction);
	    }

	    public void visitArrayLoad(IArrayLoadInstruction instruction) {
	    	// Do nothing.
	    	System.out.println("ARRAY LOAD");
	        pop();
	        pop();
	        operandStack.push(bottom());
	    	System.out.println(instruction);
	    }

	    public void visitArrayStore(IArrayStoreInstruction instruction) {
	    	// Do nothing.
	    	System.out.println("ARRAY STORE");
	    	pop();
	    	pop();
	    	pop();
	    	System.out.println(instruction);
	    }

	    public void visitPop(PopInstruction instruction) {
	    	// Do nothing.
	    	System.out.println("POP");
	    	pop();
	    	System.out.println(instruction);
	    }

	    public void visitDup(DupInstruction instruction) {
	    	System.out.println("DUP");
	    	System.out.println(operandStack);
	        int size = instruction.getSize();
	        int delta = instruction.getDelta();
	        assert size == 1 || size == 2;
	        assert delta == 0 || delta == 1 || delta == 2;
	        
	        int toPop = size + delta;
	        HashSet<Integer> v1 = pop();
	        HashSet<Integer> v2 = (toPop > 1) ? pop() : null;
	        HashSet<Integer> v3 = (toPop > 2) ? pop() : null;
	        HashSet<Integer> v4 = (toPop > 3) ? pop() : null;

	        if (size > 1) {
	          operandStack.push(v2);
	        }
	        operandStack.push(v1);
	        if (v4 != null) {
	          operandStack.push(v4);
	        }
	        if (v3 != null) {
	          operandStack.push(v3);
	        }
	        if (v2 != null) {
	          operandStack.push(v2);
	        }
	        operandStack.push(v1);
	        System.out.println(operandStack);
	    	System.out.println(instruction);
	    }

	    public void visitSwap(SwapInstruction instruction) {
	    	System.out.println("SWAP");
	    	HashSet<Integer> top1 = pop();
	    	HashSet<Integer> top2 = pop();
	    	operandStack.push(top1);
	    	operandStack.push(top2);
	    	System.out.println(instruction);
	    }

	    public void visitBinaryOp(IBinaryOpInstruction instruction) {
	    	System.out.println("BINOP");
	    	// Do nothing.
	    	pop();
	    	pop();
	    	operandStack.push(bottom());
	    	System.out.println(instruction);
	    }

	    public void visitUnaryOp(IUnaryOpInstruction instruction) {
	    	// Do nothing.
	    	System.out.println("UNARYOP");
	    	pop();
	    	operandStack.push(bottom());
	    	System.out.println(instruction);
	    }

	    public void visitShift(IShiftInstruction instruction) {
	    	System.out.println("SHIFT");
	    	pop();
	    	operandStack.push(bottom());
	    	// Do nothing.
	    	System.out.println(instruction);
	    }

	    public void visitConversion(IConversionInstruction instruction) {
	    	System.out.println("CONVERSION");
	    	// Do nothing.
	    	System.out.println(instruction);
	    }

	    public void visitComparison(IComparisonInstruction instruction) {
	    	System.out.println("COMPARISON");
	    	// Do nothing.
	    	pop();
	    	pop();
	    	operandStack.push(bottom());
	    	System.out.println(instruction);
	    }

	    public void visitConditionalBranch(IConditionalBranchInstruction instruction) {
	    	System.out.println("CONDITIONAL");
	    	// Do nothing.
	    	pop();
	    	pop();
	    	System.out.println(instruction);
	    }

	    public void visitSwitch(SwitchInstruction instruction) {
	    	System.out.println("SWITCH");
	    	// Do nothing.
	    	System.out.println(instruction);
	    }

	    public void visitReturn(ReturnInstruction instruction) {
	    	System.out.println("RETURN");
	    	// Do nothing.
	    	System.out.println(instruction);
	    }

	    public void visitGet(IGetInstruction instruction) {
	    	System.out.println("GET");
	    	HashSet<Integer> objectloc = pop();
	    	String field = instruction.getFieldName();
	    	
	    	HashSet<Integer> resultSet = new HashSet<Integer>();
	    	Iterator<Integer> it = objectloc.iterator();
	    	while (it.hasNext()) {
	    		resultSet.addAll(curState.lookupInHBar(it.next(), field));
	    	}
	    	
	    	System.out.println("Got from " + objectloc + "." + field + ": " + resultSet);
	    	operandStack.push(resultSet);

	    	System.out.println(instruction);
	    }

	    public void visitPut(IPutInstruction instruction) {
	    	HashSet<Integer> value = pop();
	    	HashSet<Integer> objectloc = pop();
	    	String field = instruction.getFieldName();
	    	
	    	Iterator<Integer> it = objectloc.iterator();
	    	System.out.println("PUT: " + objectloc + "." + field + " = " + value);
	    	while (it.hasNext()) {
	    		Integer cur = it.next();
	    		curState.setInHBar(cur, field, value);
	    	}
	    	System.out.println(instruction);
	    }

	    protected void popN(IInstruction instruction) {
	    	for (int i = 0; i < instruction.getPoppedCount(); i++) {
	    		pop();
	        }
	     }
	      
	    @Override
	    public void visitInvoke(IInvokeInstruction instruction) {
	    	System.out.println("INVOKE");
	    	popN(instruction);
	    	
	    	/*
	        ClassLoaderReference loader = cfg.getMethod().getDeclaringClass().getClassLoader().getReference();
	        TypeReference returnType = ShrikeUtil.makeTypeReference(loader, Util.getReturnType(instruction.getMethodSignature()));
	        if (!returnType.equals(TypeReference.Void)) {
	        */
	        //	operandStack.push(nullset);
	        //}
	        
	    	System.out.println(instruction);
	    }

	    public void visitNew(NewInstruction instruction) {
	    	System.out.println("NEW");
	    	popN(instruction);
	    	operandStack.push(curState.getAllocSiteSet(instruction));
	    	System.out.println("Allocation site: " + curState.getAllocSiteSet(instruction));
	    	System.out.println(instruction);
	    }

	    public void visitArrayLength(ArrayLengthInstruction instruction) {
	    	pop();
	    	operandStack.push(bottom());
	    	System.out.println(instruction);
	    }

	    public void visitThrow(ThrowInstruction instruction) {
	    	// Do nothing.
	    	System.out.println(instruction);
	    }

	    public void visitMonitor(MonitorInstruction instruction) {
	    	// Do nothing.
	    	pop();
	    	System.out.println(instruction);
	    }

	    public void visitCheckCast(CheckCastInstruction instruction) {
	    	System.out.println(instruction);
	    }

	    public void visitInstanceof(InstanceofInstruction instruction) {
	    	System.out.println(instruction);
	    }

	    public void visitLoadIndirect(ILoadIndirectInstruction instruction) {
	    	// Do nothing.
	    	System.out.println(instruction);
	    }

	    public void visitStoreIndirect(IStoreIndirectInstruction instruction) {
	    	// Do nothing.
	    	System.out.println(instruction);
	    }
	}
	
	
	static int count = 0;
	private int id=count++;
	ShrikeCFG.BasicBlock node;

	TransferFunction(ShrikeCFG.BasicBlock node){
		this.node = node;
	}
	
	void printNodeInstructions(){
		
	    BasicBlock bb = node;
	    IInstruction[] inst = null; 
	    try{
	    	inst = ((IBytecodeMethod)bb.getMethod()).getInstructions();	    
	    } catch (InvalidClassFileException e) {
	        e.printStackTrace();
	        Assertions.UNREACHABLE();	     
	    }
	    
	}
	
	@Override
	public byte evaluate(AbstractState lhs, AbstractState rhs) {
		// The lhs state is the output state for the node. The rhs state is the input state.
		// You must update the output state based on the input state and the node.
		//the return value indicates whether or not the output state changed.
		// see FixedPointConstants for the return codes.
		// printNodeInstructions();
		
		// Copy over the stack from the RHS and give it to the LHS.
		lhs.setOperandStack(rhs.getOperandStackCopy());
		System.out.println("*** BASIC BLOCK BEGIN");
		System.out.println(lhs.getOperandStack());
		System.out.println("***");
		
		boolean changed = false;
		
		// First add RHS to LHS.
		changed = changed || lhs.combineChanged(rhs);
		
		if (node.isEntryBlock()) {
			lhs.initializeHeapFields(-1);
			System.out.println("ENTRY HBAR: " + lhs.getHBar());
		}
		
		// Add everything from this node to the LHS.
		if (node.getFirstInstructionIndex() >= 0) {
			Iterator<IInstruction> it = node.iterator();
			while (it.hasNext()) {
				(it.next()).visit(new InstructionVisitor(lhs));
			}
		}
		changed = changed || !lhs.sameState(rhs);
		
		System.out.println(lhs.getOperandStack());
		System.out.println("*** BASIC BLOCK END");
		if (changed) {
			return FixedPointConstants.CHANGED;
		} else {
			System.out.println("(Not changed!)");
			return FixedPointConstants.NOT_CHANGED;
		}
	}

	@Override
	public boolean equals(Object o) {
		if (o instanceof TransferFunction) {
			return id == ((TransferFunction)o).hashCode();
		} else {
			return false;
		}
	}

	@Override
	public int hashCode() {
		return id;
	}

	@Override
	public String toString() {
		return "AbstractHeapTransferFunction";
	}
}